/** ised - generating integer and floating point sequences
    Copyright (C) 2009-2011  Simon Copar

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/

#include "base.h"

#include <cerrno>
#include <cstring>

#if IVALUE==IINT
#include <climits>
#endif

using namespace std;

#ifndef HAVE_EXP10_SET
ieee exp10(ieee x){
  return exp(LOG10*x);
}
#endif

value::value(){
  type=INTEGER;
}

value::value(itype ii){
  type=INTEGER;
  i=ii;
  d=ii;
}

itype value::intpart() const{
  if(d>=0){
    return (itype)d;
  }else{
    return (itype)(d-(itype)d+1)+(itype)d-1;
  }
}

value::value(ieee ff){
  type=FLOATING;
  d=ff;
}

value::value(const char* s){
  if(strchr(s,'.')||strchr(s,'e')||strchr(s,'E')){
    type=FLOATING;
#if VALUE==LONG
    d=strtold(s,0);
#elif VALUE==DOUBLE
    d=strtod(s,0);
#else
    d=strtof(s,0);
#endif
    return;
  }else{
    type=INTEGER;
#if IVALUE==ILONG
    errno=0;
    long long int ival=strtoll(s,0,10);
    if(errno){
      fprintf(stderr,"parse: integer overflow during parsing input\n");
      type=FLOATING;
#ifdef NAN
      d=NAN;
#else
      d=0./0.;
#endif
      i=-1;
    }
    else{
      i=ival;
      d=i;
    }

#else
    errno=0;
    long int ival=strtol(s,0,10);
    if(errno || ival>INT_MAX || ival<INT_MIN){
      fprintf(stderr,"parse: integer overflow during parsing input\n");
      type=FLOATING;
#ifdef NAN
      d=NAN;
#else
      d=0./0.;
#endif
      i=-1;
    }else{
      i=ival;
      d=i;
    }

#endif
  }
}

itype value::assertI(const char* op){
  if(type!=INTEGER){
    fprintf(stderr,"warning: integer required by '%s', converting\n",op);
    return intpart();
  }
  return i;
}

ieee value::assertD(const char* op){
  if(type!=FLOATING){
    fprintf(stderr,"warning: floating required by '%s', converting\n",op);
  }
  return d;
}


//assembler routines for trapping overflows
#ifdef TRAP_OVR
inline static void iadd_(itype& a, itype b, char& flag){
    char fl;
#if IVALUE==IINT
    __asm__ __volatile__ (
			"addl %2,%0\n\t"
			"seto %1\n"
			:"=g"(a), "=ir"(fl)
			:"ir"(b), "0"(a)
			);
#elif IVALUE==ILONG
    __asm__ __volatile__ (
			"addq %2,%0\n\t"
			"seto %1\n"
			:"=g"(a), "=ir"(fl)
			:"ir"(b), "0"(a)
			);
#endif
    flag|=fl;
}

inline static void isub_(itype& a, itype b, char& flag){
    char fl;
#if IVALUE==IINT
    __asm__ __volatile__ (
			"subl %2,%0\n\t"
			"seto %1\n"
			:"=g"(a), "=ir"(fl)
			:"ir"(b), "0"(a)
			);
#elif IVALUE==ILONG
    __asm__ __volatile__ (
			"subq %2,%0\n\t"
			"seto %1\n"
			:"=g"(a), "=ir"(fl)
			:"ir"(b), "0"(a)
			);
#endif
    flag|=fl;
}

inline static void imul_(itype& a, itype b, char& flag){
    char fl;
#if IVALUE==IINT
    __asm__ __volatile__ (
			"imull %2,%0\n\t"
			"seto %1\n"
			:"=g"(a), "=ir"(fl)
			:"ir"(b), "0"(a)
			);
#elif IVALUE==ILONG
    __asm__ __volatile__ (
			"imulq %2,%0\n\t"
			"seto %1\n"
			:"=g"(a), "=ir"(fl)
			:"ir"(b), "0"(a)
			);
#endif
    flag|=fl;
}
#endif
//end of assembler routines



value value::operator+(const value& v) const{
  if((type==INTEGER)&&(v.type==INTEGER)){

#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v.i;
    iadd_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on addition, casting to float\n");
#endif
      return value(d+v.d);
    }else{
      return value(result);
    }

#else
    return value(i+v.i);
#endif
  }else{
    return value(d+v.d);
  }
}

value value::operator-(const value& v) const{
  if((type==INTEGER)&&(v.type==INTEGER)){

#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v.i;
    isub_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on subtraction, casting to float\n");
#endif
      return value(d-v.d);
    }else{
      return value(result);
    }

#else
    return value(i-v.i);
#endif
  }else{
    return value(d-v.d);
  }
}


value value::operator*(const value& v) const{
  if((type==INTEGER)&&(v.type==INTEGER)){
#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v.i;
    imul_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on multiplication, casting to float\n");
#endif
      return value(d*v.d);
    }else{
      return value(result);
    }

#else
    return value(i*v.i);
#endif
  }else{
    return value(d*v.d);
  }
}

value value::operator/(const value& v) const{
  if((type==INTEGER)&&(v.type==INTEGER)){
    if(v.i==0)return value((itype)0);
    else return value(i/v.i);
  }else{
    if(v.d==0)return value((itype)0);
    else return value(d/v.d);
  }
}

static value flpow(ieee i1, itype i2){
  if(i2<0){
    return value(pow(i1,(int)i2));
  }else{
    int sign=(i1<0) && (i2&1);
    i1=i1<0?-i1:i1;
    ieee trailing=i1;
    ieee total=1;
    if(i2&1)total*=trailing;
    i2>>=1;
    while(i2!=0){
      trailing*=trailing;
      if(i2&1)total*=trailing;
      i2>>=1;
    }
    if(sign)total=-total;
    return value(total);
  }
}

inline static itype imul(itype a, itype b, char& flag){
#ifdef TRAP_OVR
  imul_(a,b,flag);
#if DEBUG>0
  if(flag){
    fprintf(stderr,"runtime: integer overflow on multiplication\n");
  }
#endif
  return a;  
#else
  return a*b;
#endif
}


static value intpow(itype i1, itype i2){
  if(i2<0){
    return value(pow((ieee)i1,(int)i2));
  }else{
    char flag=0;
    int sign=(i1<0) && (i2&1);
    itype base=i1<0?-i1:i1;
    itype trailing=base;
    itype total=1;
    itype ex=i2;
    if(ex&1)total=imul(total,trailing,flag);
    ex>>=1;
    while(ex!=0){
      trailing=imul(trailing,trailing,flag);
      if(ex&1)total=imul(total,trailing,flag);
      ex>>=1;
    }
    if(flag)return flpow((ieee)i1,i2);
    if(sign)total=-total;
    return value((itype)total);
  }
}

value value::operator^(const value& v) const{
  if((type==INTEGER)&&(v.type==INTEGER)){
    return intpow(i,v.i);
  }else if(v.type==INTEGER){
    return flpow(d,v.i);
  }else{
    return value(pow(d,v.d));
  }
}

static ieee d_mod(ieee d1, ieee d2){
  ieee r=d1-(floor(d1/d2))*d2;
  return r<0?0:r;
}

static ieee f_mod(ieee v1, ieee v2){
  if(v2==0)return v1;
  if(v2>=0){
    return d_mod(v1,v2);
  }else{
    v2=-v2;
    return d_mod(v1,v2)-v2;
  }
}

static itype i_mod(itype i1, itype i2){
  if(i2==0)return i1;
  if(i2>=0){
    //return i1<0?(i2-(-i1)%i2)%i2:i1%i2;//room for improvement
    if(i1<0){
      itype it=(i2-(-i1)%i2);
      if(it==i2)return 0;
      else return it;
    }else{
      return i1%i2;
    }
  }else{
    i2=-i2;
    itype it;
    if(i1<0){
      it=(i2-(-i1)%i2);
      if(it==i2)it=0;
    }else{
      it=i1%i2;
    }
    return it-i2;
    //return (i1<0?(i2-(-i1)%i2)%i2:i1%i2)-i2;
  }
}

value value::operator%(const value& v) const{
  if((type==INTEGER)&&(v.type==INTEGER)){
    return value(i_mod(i,v.i));
  }else{
    return value(f_mod(d,v.d));
  }
}

value value::e(const value& v) const {
  return value(d*exp10(v.d));
}

value value::binomial(const value& v) const {
  if(v.type==INTEGER){
    value total=(itype)1;
    value nom=(*this);
    for(int i=1;i<=v.i;i++){
      total*=nom;
      total=total/(itype)i;
      nom=nom-(itype)1;
    }
    return total;
  }else{
    int i1, i2, i3;
    ieee g1=lgamma_r(d+1,&i1);
    ieee g2=lgamma_r(v.d+1,&i2);
    ieee g3=lgamma_r(d-v.d+1,&i3);
    return value((ieee)((i1*i2*i3)*exp(g1-g2-g3)));
  }
}



void value::operator+=(const value& v){
  if((type==INTEGER)&&(v.type==INTEGER)){

#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v.i;
    iadd_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on addition, casting to float\n");
#endif
      type=FLOATING;
      d+=v.d;
    }else{
      i=result;
      d=i;
    }

#else
    i+=v.i;
    d=i;
#endif
  }else{
    type=FLOATING;
    d+=v.d;
  }
}

void value::operator-=(const value& v){
  if((type==INTEGER)&&(v.type==INTEGER)){

#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v.i;
    isub_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on subtraction, casting to float\n");
#endif
      type=FLOATING;
      d-=v.d;
    }else{
      i=result;
      d=i;
    }

#else
    i-=v.i;
    d=i;
#endif
  }else{
    type=FLOATING;
    d-=v.d;
  }
}


void value::operator*=(const value& v){
  if((type==INTEGER)&&(v.type==INTEGER)){
#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v.i;
    imul_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on multiplication, casting to float\n");
#endif
      type=FLOATING;
      d*=v.d;
    }else{
      i=result;
      d=i;
    }

#else
    i*=v.i;
    d=i;
#endif
  }else{
    type=FLOATING;
    d*=v.d;
  }
}

void value::operator/=(const value& v){
  if((type==INTEGER)&&(v.type==INTEGER)){
    if(v.i==0){
      i=0;
      d=0;
    }else{
      i/=v.i;
      d=i;
    }
  }else{
    if(v.d==0){
      type=INTEGER;
      i=0;
      d=0;      
    }else{
      type=FLOATING;
      d/=v.d;
    }
  }
}



value value::operator+(const ieee& v) const{
  return value(d+v);
}

value value::operator-(const ieee& v) const{
  return value(d-v);
}

value value::operator*(const ieee& v) const{
  return value(d*v);
}

value value::operator/(const ieee& v) const{
  if(v==0)return value((itype)0);
  return value(d/v);
}

value value::operator^(const ieee& v) const{
  return value(pow(d,v));
}

value value::operator%(const ieee& v) const{
  return value(f_mod(d,v));
}


value value::operator+(const itype& v) const{
  if(type==INTEGER){
#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v;
    iadd_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on addition, casting to float\n");
#endif
      return value(d+(ieee)v);
    }else{
      return value(result);
    }

#else
    return value(i+v);
#endif
  }else{
    return value(d+v);
  }
}

value value::operator-(const itype& v) const{
  if(type==INTEGER){
#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v;
    isub_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on subtraction, casting to float\n");
#endif
      return value(d-(ieee)v);
    }else{
      return value(result);
    }

#else
    return value(i-v);
#endif
  }else{
    return value(d-v);
  }
}

value value::operator*(const itype& v) const{
  if(type==INTEGER){
#ifdef TRAP_OVR
    char flag;
    itype result=i;
    itype operand=v;

    imul_(result,operand,flag);
    if(flag){
#if DEBUG>0
      fprintf(stderr,"runtime: integer overflow on multiplication, casting to float\n");
#endif
      return value(d*(ieee)v);
    }else{
      return value(result);
    }

#else
    return value(i*v);
#endif
  }else{
    return value(d*v);
  }
}

value value::operator/(const itype& v) const{
  if(v==0)return value((itype)0);
  if(type==INTEGER){
    return value(i/v);
  }else{
    return value(d/v);
  }
}

value value::operator^(const itype& v) const{
  if(type==INTEGER){
    return intpow(i,v);
  }else {
    return flpow(d,v);
  }
}

value value::operator%(const itype& v) const{
  if(type==INTEGER){
    return value(i_mod(i,v));
  }else{
    return value(f_mod(d,v));
  }
}

value value::operator-() const{
  if(type==INTEGER){
    return value(-i);
  }else{
    return value(-d);
  }
}

void value::op_neg(){
  i=-i;
  d=-d;
  
}

void value::op_sin(){
  d=sin(d);
  type=FLOATING;
}

void value::op_cos(){
  d=cos(d);
  type=FLOATING;
}

void value::op_tan(){
  d=tan(d);
  type=FLOATING;
}

void value::op_atan(){
  d=atan(d);
  type=FLOATING;
}

void value::op_asin(){
  d=asin(d);
  type=FLOATING;
}

void value::op_acos(){
  d=acos(d);
  type=FLOATING;
}

void value::op_abs(){
  if(type==INTEGER){
    if(i<0){i=-i;d=-d;}
  }else{
    if(d<0)d=-d;
  }
}

void value::op_exp(){
  d=exp(d);
  type=FLOATING;
}

void value::op_sqrt(){
  d=sqrt(d);
  type=FLOATING;
}

void value::op_cbrt(){
  d=cbrt(d);
  type=FLOATING;
}

void value::op_log(){
  d=log(d);
  type=FLOATING;
}


value value::op_bj(int n){
  return value((ieee)jn(n,d));
}

value value::op_by(int n){
  return value((ieee)yn(n,d));
}

void value::op_int(){
  if(type==FLOATING){
    type=INTEGER;
    i=intpart();
    d=i;
  }
}

void value::op_float(){
  type=FLOATING;
}

void value::op_frac(){
  if(type==FLOATING){
   itype ipart=intpart();
   d-=ipart;
  }else{
    type=FLOATING;
    d=0;
  }
}

void value::op_round(){
  if(type==FLOATING){
   i=intpart();
   d-=i;
   if(d>=0.5)i++;
   type=INTEGER;
   d=i;
  }
}


#if IVALUE==ILONG
#define MAX_FACT_INT 20
#else
#define MAX_FACT_INT 12
#endif

void value::op_fact(){
  if(type==INTEGER){
    if(i>MAX_FACT_INT){
      d=1;
      for(itype u=2;u<=i;u++)d*=u;
      type=FLOATING;
    }else{
      itype fac=1;
      for(itype u=2;u<=i;u++)fac*=u;
      d=i=fac;
    }
  }else{
    int gi;
    ieee lg=lgamma_r(d+1,&gi);
    d=gi*exp(lg);
  }
}

void value::op_exp10(){
  d=exp10(d);
  type=FLOATING;
}

void value::op_rand(){
  if(type==INTEGER){
    if(!i)return;
    int sign=i<0?-1:1;
    i=sign*(rand()%(sign*i));
    d=i;
  }else{
    d=(d*rand())/RAND_MAX;
  }
}

int value::is_odd() const{
  if(type!=INTEGER)return 0;
  return i%2!=0;
}

int value::is_even() const{
  if(type!=INTEGER)return 0;
  return i%2==0;
}

int value::is_positive() const{
  if(type==INTEGER)return i>=0;
  else return d>=0;
}

int value::is_negative() const{
  if(type==INTEGER)return i<0;
  else return d<0;
}

int value::is_int() const{
  return type==INTEGER;
}

int value::is_float() const{
  return type==FLOATING;
}

int operator<(const value& v1, const value& v2){
  return v1.d<v2.d;
}

int operator>(const value& v1, const value& v2){
  return v1.d>v2.d;
}

int operator<=(const value& v1, const value& v2){
  return v1.d<=v2.d;
}

int operator>=(const value& v1, const value& v2){
  return v1.d>=v2.d;
}

int operator==(const value& v1, const value& v2){
  if(v1.type!=v2.type)return 0;
  if(v1.type==INTEGER)return v1.i==v2.i;
  else return v1.d==v2.d;
}

int operator!=(const value& v1, const value& v2){
  if(v1.type!=v2.type)return 1;
  if(v1.type==INTEGER)return v1.i!=v2.i;
  else return v1.d!=v2.d;
}

value value::lt(const value& v) const{
  return value((itype)(d<v.d));
}

value value::gt(const value& v) const{
  return value((itype)(d>v.d));
}

value value::le(const value& v) const{
  return value((itype)(d<=v.d));
}

value value::ge(const value& v) const{
  return value((itype)(d>=v.d));
}

value value::eq(const value& v) const{
  if(type!=v.type)return value((itype)0);
  if(type==INTEGER)return value((itype)(i==v.i));
  else return value((itype)(d==v.d));
}

value value::neq(const value& v) const{
  if(type!=v.type)return value((itype)1);
  if(type==INTEGER)return value((itype)(i!=v.i));
  else return value((itype)(d!=v.d));
}


void value::print(FILE* f, const char* Fd, const char* Ff){
  switch(type){
    case INTEGER:
      fprintf(f,Fd,i);
      return;
    case FLOATING:
      fprintf(f,Ff,d);
      return;
  }
}

//definition of shared::shared_ops, declared externally in the header
opid shared::shared_ops[SHARED_OPS_LEN]={
  //parentheses operators
  (opid){("{}"),2,OP_LIST     ,0,ASOC_MULTI},   //0
  (opid){("[]"),2,OP_ITER1    ,0,ASOC_MULTI},   //1
  (opid){("()"),2,OP_ITER2    ,0,ASOC_MULTI},   //2
  (opid){("{::}"),4,OP_FUN    ,0,ASOC_MULTI},   //3
  //special case
  (opid){("-"),1,OP_MINUSSIGN ,11,ASOC_RUNARY}, //4

  (opid){("$"),1,OP_TAKE      ,9,ASOC_RUNARY},  //5
  (opid){("@"),1,OP_PUT       ,9,ASOC_RBINARY}, //6
  (opid){("Q"),1,OP_SYSCALL   ,9,ASOC_RUNARY},  //7
  (opid){("::"),2,OP_FCALL    ,8,ASOC_RINFIX},  //8
  (opid){(":::"),3,OP_FITER   ,8,ASOC_RINFIX},  //9
  (opid){("x"),1,OP_X         ,11,ASOC_BLANK},  //10

  (opid){(":+"),2,OP_PPLUS    ,4,ASOC_LINFIX},  //11
  (opid){(":-"),2,OP_MMINUS   ,4,ASOC_LINFIX},  //12
  (opid){(":*"),2,OP_MMUL     ,5,ASOC_LINFIX},  //13
  (opid){(":/"),2,OP_DDIV     ,5,ASOC_LINFIX},  //14
  (opid){(":%"),2,OP_MMOD     ,5,ASOC_LINFIX},  //15
  (opid){(":^"),2,OP_PPOW     ,6,ASOC_RINFIX},  //16

  (opid){("@+"),2,OP_OPLUS    ,8,ASOC_RUNARY},  //17
  (opid){("@*"),2,OP_OTIMES   ,8,ASOC_RUNARY},  //18
  (opid){("avg"),3,OP_OMEAN   ,8,ASOC_RUNARY},  //19
  (opid){("var"),3,OP_OVAR    ,8,ASOC_RUNARY},  //20
  (opid){("nrm"),3,OP_ONORM   ,8,ASOC_RBINARY}, //21

  (opid){("!"),1,OP_FACT      ,9,ASOC_LUNARY},  //22
  (opid){("sin"),3,OP_SIN     ,8,ASOC_RUNARY},  //23
  (opid){("cos"),3,OP_COS     ,8,ASOC_RUNARY},  //24
  (opid){("tan"),3,OP_TAN     ,8,ASOC_RUNARY},  //25
  (opid){("atn"),3,OP_ATAN    ,8,ASOC_RUNARY},  //26
  (opid){("atan"),4,OP_ATAN   ,8,ASOC_RUNARY},  //27
  (opid){("asn"),3,OP_ASIN    ,8,ASOC_RUNARY},  //28
  (opid){("asin"),4,OP_ASIN   ,8,ASOC_RUNARY},  //29
  (opid){("acs"),3,OP_ACOS    ,8,ASOC_RUNARY},  //30
  (opid){("acos"),4,OP_ACOS   ,8,ASOC_RUNARY},  //31
  (opid){("abs"),3,OP_ABS     ,8,ASOC_RUNARY},  //32
  (opid){("exp"),3,OP_EXP     ,8,ASOC_RUNARY},  //33
  (opid){("sqt"),3,OP_SQRT    ,8,ASOC_RUNARY},  //34
  (opid){("sqrt"),4,OP_SQRT   ,8,ASOC_RUNARY},  //35
  (opid){("cbt"),3,OP_CBRT    ,8,ASOC_RUNARY},  //36
  (opid){("cbrt"),4,OP_CBRT   ,8,ASOC_RUNARY},  //37
  (opid){("r"),1,OP_ROUND     ,8,ASOC_RUNARY},  //38
  (opid){("f"),1,OP_FRAC      ,8,ASOC_RUNARY},  //39
  (opid){("i"),1,OP_INT       ,8,ASOC_RUNARY},  //40
  (opid){("d"),1,OP_FLOAT     ,8,ASOC_RUNARY},  //41
  (opid){("log"),3,OP_LOG     ,8,ASOC_RUNARY},  //42
  (opid){("ran"),3,OP_RAND    ,8,ASOC_RUNARY},  //43
  (opid){("bj"),2,OP_BESJ     ,8,ASOC_RBINARY}, //44
  (opid){("by"),2,OP_BESY     ,8,ASOC_RBINARY}, //45

  (opid){("+"),1,OP_PLUS      ,4,ASOC_LINFIX},  //46
  (opid){("-"),1,OP_MINUS     ,4,ASOC_LINFIX},  //47
  (opid){("*"),1,OP_MUL       ,5,ASOC_LINFIX},  //48
  (opid){("/"),1,OP_DIV       ,5,ASOC_LINFIX},  //49
  (opid){("%"),1,OP_MOD       ,5,ASOC_LINFIX},  //50
  (opid){("^"),1,OP_POW       ,6,ASOC_RINFIX},  //51
  (opid){("e"),1,OP_FLE       ,10,ASOC_RINFIX}, //52
  (opid){("c"),1,OP_NCK       ,7,ASOC_RINFIX},  //53

  (opid){("&"),1,OP_AND       ,3,ASOC_LINFIX},  //54
  (opid){("\\"),1,OP_WITHOUT  ,3,ASOC_LINFIX},  //55
  (opid){("|"),1,OP_OR        ,2,ASOC_LINFIX},  //56
  (opid){("|&"),2,OP_XOR      ,3,ASOC_LINFIX},  //57
  (opid){("~"),1,OP_REVERSE   ,8,ASOC_RUNARY},  //58
  (opid){("S"),1,OP_SORT      ,8,ASOC_RUNARY},  //59
  (opid){("#"),1,OP_HASH      ,8,ASOC_RUNARY},  //60
  (opid){("?"),1,OP_MASK      ,8,ASOC_RUNARY},  //61
  (opid){("="),1,OP_EQUALS    ,1,ASOC_LINFIX},  //62
  (opid){("#_"),2,OP_COUNT    ,8,ASOC_RBINARY}, //63
  (opid){("#__"),3,OP_HIST    ,8,ASOC_RBINARY}, //64

  (opid){("<<"),2,OP_SHL      ,7,ASOC_LINFIX},  //65
  (opid){(">>"),2,OP_SHR      ,7,ASOC_LINFIX},  //66
  (opid){("_"),1,OP_INDEX     ,7,ASOC_LINFIX},  //67
  (opid){("U"),1,OP_UNIQUE    ,8,ASOC_RUNARY},  //68
  (opid){("P"),1,OP_POSITIVE  ,8,ASOC_RUNARY},  //69
  (opid){("N"),1,OP_NEGATIVE  ,8,ASOC_RUNARY},  //70
  (opid){("O"),1,OP_ODD       ,8,ASOC_RUNARY},  //71
  (opid){("E"),1,OP_EVEN      ,8,ASOC_RUNARY},  //72
  (opid){("I"),1,OP_INTEGER   ,8,ASOC_RUNARY},  //73
  (opid){("D"),1,OP_DOUBLE    ,8,ASOC_RUNARY},  //74
  (opid){("min"),3,OP_MIN     ,8,ASOC_RUNARY},  //75
  (opid){("max"),3,OP_MAX     ,8,ASOC_RUNARY},  //76

  (opid){(":<"),2,OP_LLT      ,1,ASOC_LINFIX},  //77
  (opid){(":>"),2,OP_GGT      ,1,ASOC_LINFIX},  //78
  (opid){(":<="),3,OP_LLEQ    ,1,ASOC_LINFIX},  //79
  (opid){(":>="),3,OP_GGEQ    ,1,ASOC_LINFIX},  //80
  (opid){(":=="),3,OP_EEQ     ,1,ASOC_LINFIX},  //81
  (opid){(":<>"),3,OP_NNEQ    ,1,ASOC_LINFIX},  //82

  (opid){("<"),1,OP_LT        ,1,ASOC_LINFIX},  //83
  (opid){(">"),1,OP_GT        ,1,ASOC_LINFIX},  //84
  (opid){("<="),2,OP_LEQ      ,1,ASOC_LINFIX},  //85
  (opid){(">="),2,OP_GEQ      ,1,ASOC_LINFIX},  //86
  (opid){("=="),2,OP_EQ       ,1,ASOC_LINFIX},  //87
  (opid){("<>"),2,OP_NEQ      ,1,ASOC_LINFIX},  //88

  (opid){("pi"),2,OP_PI       ,11,ASOC_BLANK},  //89
  (opid){("aa"),2,OP_AA       ,11,ASOC_BLANK},  //90
  (opid){("deg"),3,OP_DEG     ,11,ASOC_BLANK},  //91
  (opid){("emc"),3,OP_EMC     ,11,ASOC_BLANK},  //92

  (opid){("pm"),2,OP_PM       ,11,ASOC_BLANK},  //93
  (opid){("mp"),2,OP_MP       ,11,ASOC_BLANK},  //94

  (opid){("++"),2,OP_XPLUS    ,4,ASOC_LINFIX},  //95
  (opid){("--"),2,OP_XMINUS   ,4,ASOC_LINFIX},  //96
  (opid){("**"),2,OP_XMUL     ,5,ASOC_LINFIX},  //97
  (opid){("//"),2,OP_XDIV     ,5,ASOC_LINFIX},  //98
  (opid){("%%"),2,OP_XMOD     ,5,ASOC_LINFIX},  //99
  //(opid){("^^"),2,OP_XPOW     ,6,ASOC_RINFIX},
  (opid){("px"),2,OP_PX       ,8,ASOC_RBINARY}, //100
  (opid){("pd"),2,OP_PD       ,8,ASOC_RUNARY},  //101
  (opid){("pz"),2,OP_PZ       ,8,ASOC_RUNARY},  //102

  (opid){("@:"),2,OP_DOT      ,5,ASOC_LINFIX},  //103
  (opid){("@="),2,OP_VLEN     ,8,ASOC_RUNARY},  //104
  (opid){("@!"),2,OP_VNRM     ,8,ASOC_RUNARY},  //105
  (opid){("@^"),2,OP_CROSS    ,6,ASOC_RINFIX}   //106
};

#ifndef NOSYSCALL
int shared::special_interrupt=INTERRUPT_NONE;
#endif
int shared::exit_returnvalue=0;

#ifdef SIGHANDLE
sig_atomic_t shared::interactive_flag=0;
sig_atomic_t shared::state=STATE_IDLE;

sig_atomic_t shared::interrupt=0; //volatile?
char* shared::wait_ptr=0;
#endif

opid findop(char* c){
  int idx=-1;
  uint maxlen=0;
  for(int i=5;i<SHARED_OPS_LEN;i++){
    uint len=shared::shared_ops[i].len;
    if(len<=maxlen)continue;
    if(!strncmp(shared::shared_ops[i].name,c,len)){
      maxlen=len;
      idx=i;
    }
  }
  if(idx<0)return (opid){0,0,0,0};
  return shared::shared_ops[idx];
}
